"""
.. module:: date
   :platform: Unix, Windows
   :synopsis: Module to abstract the date based data features for wunderground

.. moduleauthor:: Tyler Santos <1tsantos at gmail.com>

"""
__docformat__ = 'reStructuredText'

# local Ms
#from weather_base import WeatherBase
from WunderWeather.weather_base import WeatherBase


class Date(WeatherBase):
    """
    Wrapper for one (date) history type data feature response.
    """

    def __init__(self, data, *args, **kwargs):
        super(Date, self).__init__(data, *args, **kwargs)
        """constructor to interface with feature response

        Args:
            data (dict): JSON response

        """
        pass

    @property
    def temp_f(self):
        return self.extract_value(['dailysummary', 0, 'meantempi'])

    @property
    def temp_c(self):
        return self.extract_value(['dailysummary', 0, 'meantempm'])

    @property
    def observations(self):
        """Abstract the observations for given date

        Notes:
            In a date based response (History,Yesterday) there is a list of observations.

        Attributes:
            observations (list): List of dictionaries for each observation for the date

        Returns:
            list of Observation instances

        """
        observations = self.extract_value(['observations'])
        if observations:
            return [Observation(data=obs) for obs in observations]


class Observation(WeatherBase):
    """
    Wrapper for one date based data feature's observations
    """

    def __init__(self, data, *args, **kwargs):
        super(Observation, self).__init__(data, *args, **kwargs)
        """constructor to interface with feature response for one observation

        Args:
            data (dict): JSON response

        """
        pass

    @property
    def temp_f(self):
        return self.extract_value(['tempi'])

    @property
    def temp_c(self):
        return self.extract_value(['tempm'])

    @property
    def date_pretty(self):
        return self.extract_value(['date', 'pretty'])


class Range(WeatherBase):
    """
    Wrapper for one (date) history type data feature response.
    """

    def __init__(self, data, *args, **kwargs):
        super(Range, self).__init__(data, *args, **kwargs)
        """constructor to interface with feature response

        Args:
            data (dict): JSON response

        """
        pass

    @property
    def high_avg_temp_f(self):
        return self.extract_value(['temp_high', 'avg', 'F'])

    @property
    def high_avg_temp_c(self):
        return self.extract_value(['temp_high', 'avg', 'C'])

    @property
    def low_avg_temp_f(self):
        return self.extract_value(['temp_low', 'avg', 'F'])

    @property
    def low_avg_temp_c(self):
        return self.extract_value(['temp_low', 'avg', 'C'])



"""
.. module:: test
   :platform: Unix, Windows
   :synopsis: Module to run tests for features

.. moduleauthor:: Tyler Santos <1tsantos at gmail.com>

"""
__docformat__ = 'reStructuredText'

# builtin P&Ms
import unittest
from pprint import pprint

# TODO: allow for API key to be passed in or Env Var
api_key = ''


def run_test():

    #w = weather.Extract(api_key,settings={"lang":"FR"})
    w = weather.Extract(api_key)
    # This willa actually run the connx
    #now = w.for_today_now(['MA','Boston'])
    #now = w.today_now('02481')

    # this will use cached version for testing
    #feature_key = weather.Extract.FEATURE_RESPONSE_MAP['now']
    #now = today.Now(data=today.now_response[feature_key])
    # print("TEMP:",now.temp_f)
    # print(type(now.temp_f))
    # print("Nonshort:",now.data.display_location.full)
    # pprint(now.data)

    # feature_key = weather.Extract.FEATURE_RESPONSE_MAP['date']
    # d = date.Date(data=date.date_response[feature_key])
    # print("TEMP:",d.temp_f)

    # # print all observations for the date specified
    # [print(obs.date_pretty,obs.temp_f,obs.heat_index_f) for obs in d.observations]

    # make a composite features request
    query = ['MA','Boston']
    feature_context = (
                        ('now',''),
                        ('today_historical',''),
                        ('date',"20161111"),
                        )
    [now,history,date] = w.features(query,feature_context)
    print(now)
    print(history)
    print(date)

    # this will use cached version for testing
    #feature_key = weather.Extract.FEATURE_RESPONSE_MAP['daycast']
    #feature_key = weather.Extract.FEATURE_RESPONSE_MAP['weekcast']

    # print(feature_key)
    # pprint(forecast.forecast_10_day_response)
    # print('\n\n')
    #response = forecast.daycast_response[feature_key]
    #response = forecast.forecast_10_day_response[feature_key]
    #futurecast = forecast.Forecast(data=response)
    # print("DATE:",futurecast.date)

    # for period in futurecast.periods:
    # print(period.period)
    # pprint(period.data)
    # print(period.date_pretty)
    # print(period.data.date.pretty)
    # print('\n\n')

    # Testing hurricane logic
    #hurricane = w.hurricane()
    # pprint(hurricane.data)


class TestAlerts(unittest.TestCase):

    def setUp(self):

      # feature specific vars
        feature_key = 'alerts'
        response = test_responses.alerts_response

        # setup for feature
        resp_feature_key = Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = Extract.get_feature_class(feature_key)

        # create response
        self.test_alerts = ctor({resp_feature_key: response[resp_feature_key]})

    def test_get_item_count(self):

        self.assertEqual(len(self.test_alerts.data.alerts), 1,
                         'Incorrect number of items')

    def test_get_one_item_data(self):

        self.assertEqual(self.test_alerts.data.alerts[0].description, 'Heat Advisory',
                         'Incorrect data value')


class TestTodayHistorical(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'today_historical'
        response = test_responses.historical_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_today_historical = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(self.test_today_historical.data.temp_high.normal.F, '71',
                         'data based gets are broken')

    def test_shorthand_based_get(self):

        self.assertEqual(self.test_today_historical.high_avg_temp_f, '71',
                         'shorhand gets are broken')


class TestAstronomy(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'astronomy'
        response = test_responses.astronomy_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_astronomy = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(self.test_astronomy.data.sunrise.hour, '7',
                         'data based gets are broken')


class TestTodayNow(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'now'
        response = test_responses.now_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response obj instance
        self.test_today_now = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(self.test_today_now.data.display_location.full, 'Boston, MA',
                         'data based gets are broken')

    def test_shorthand_based_get(self):

        self.assertEqual(self.test_today_now.temp_f, 47.1,
                         'shorhand gets are broken')


class TestHurricane(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'hurricane'
        response = test_responses.hurricane_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_hurricane = ctor(
            {resp_feature_key: response[resp_feature_key]})

    def test_data_based_get(self):

        self.assertEqual(self.test_hurricane.data.currenthurricane[0].stormInfo.stormName_Nice, "Hurricane Daniel",
                         'data based gets are broken')


class TestDaycast(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'daycast'
        response = test_responses.daycast_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_daycast = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(len(self.test_daycast.data.txt_forecast.forecastday), 8,
                         'data based gets are broken')

    def test_shorthand_based_get(self):
        self.assertEqual(len(self.test_daycast.periods), 8,
                         'shorhand gets are broken')

    def test_shorthand_child_based_get(self):

        period = self.test_daycast.periods[1]
        self.assertEqual(period.high_temp_f, "68",
                         'shorhand child gets are broken')


class TestWeekcast(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'weekcast'
        response = test_responses.weekcast_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_weekcast = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(len(self.test_weekcast.data.txt_forecast.forecastday), 20,
                         'data based gets are broken')

    def test_shorthand_based_get(self):
        self.assertEqual(len(self.test_weekcast.periods), 20,
                         'shorhand gets are broken')

    def test_shorthand_child_based_get(self):

        period = self.test_weekcast.periods[1]
        self.assertEqual(period.high_temp_f, "75",
                         'shorhand child gets are broken')


class TestGeolookup(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'geolookup'
        response = test_responses.geolookup_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_geolookup = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(len(self.test_geolookup.data.nearby_weather_stations), 2,
                         'data based gets are broken')


class TestDate(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'date'
        response = test_responses.date_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_date = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(len(self.test_date.data.observations), 25,
                         'data based gets are broken')

    def test_shorthand_based_get(self):
        self.assertEqual(len(self.test_date.observations), 25,
                         'shorhand gets are broken')

    def test_shorthand_child_based_get(self):

        observation = self.test_date.observations[1]
        self.assertEqual(observation.temp_f, "39.9",
                         'shorhand child gets are broken')


class TestHourlyDaycast(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'hourly_daycast'
        response = test_responses.hourly_daycast_response

        # setup for feature
        resp_feature_key = Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = Extract.get_feature_class(feature_key)

        # create response
        self.test_hourly_daycast = ctor(
            {resp_feature_key: response[resp_feature_key]})

    def test_get_item_count(self):

        self.assertEqual(len(self.test_hourly_daycast.data.hourly_forecast), 36,
                         'Incorrect number of items')

    def test_get_one_item_data(self):

        self.assertEqual(self.test_hourly_daycast.data.hourly_forecast[0].temp.english, '66',
                         'Incorrect data value')


class TestRange(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'date_range'
        response = test_responses.date_range_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_date_range = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(self.test_date_range.data.temp_high.avg.F, '71',
                         'data based gets are broken')

    def test_shorthand_based_get(self):
        self.assertEqual(self.test_date_range.high_avg_temp_f, '71',
                         'shorhand gets are broken')


class TestRawtide(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'rawtide'
        response = test_responses.rawtide_response

        # setup for feature
        resp_feature_key = Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = Extract.get_feature_class(feature_key)

        # create response
        self.test_rawtide = ctor(response[resp_feature_key])

    def test_get_item_count(self):

        self.assertEqual(len(self.test_rawtide.data.tideInfo), 1,
                         'Incorrect number of items')

    def test_get_one_item_data(self):

        self.assertEqual(self.test_rawtide.data.tideInfo[0].tideSite,
                         "Newport Beach, Newport Bay Entrance, Corona del Mar, California",
                         'Incorrect data value')


class TestTide(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'tide'
        response = test_responses.tide_response

        # setup for feature
        resp_feature_key = Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = Extract.get_feature_class(feature_key)

        # create response
        self.test_tide = ctor(response[resp_feature_key])

    def test_get_item_count(self):

        self.assertEqual(len(self.test_tide.data.tideInfo), 1,
                         'Incorrect number of items')

    def test_get_one_item_data(self):

        self.assertEqual(self.test_tide.data.tideSummary[0].data.type,
                         "Full Moon",
                         'Incorrect data value')


class TestYesterday(unittest.TestCase):

    def setUp(self):

        # feature specific vars
        feature_key = 'yesterday'
        response = test_responses.yesterday_response

        # setup for feature
        resp_feature_key = weather.Extract.FEATURE_RESPONSE_MAP[feature_key]
        ctor = weather.Extract.get_feature_class(feature_key)

        # create response
        self.test_yesterday = ctor(response[resp_feature_key])

    def test_data_based_get(self):

        self.assertEqual(len(self.test_yesterday.data.observations), 26,
                         'data based gets are broken')

    def test_shorthand_based_get(self):
        self.assertEqual(len(self.test_yesterday.observations), 26,
                         'shorhand gets are broken')

    def test_shorthand_child_based_get(self):

        observation = self.test_yesterday.observations[1]
        self.assertEqual(observation.temp_f, "59.0",
                         'shorhand child gets are broken')

def run_snippet():
    pass

def main():
    run_test()
    #run_snippet()


if __name__ == '__main__':
    # main()
    import os
    import sys
    sys.path.insert(0, os.path.abspath('../'))

    #local Ms
    import test_responses
    import weather
    from weather import *
    
    #unittest.main()
    main()



"""
.. module:: weather_base
   :platform: Unix, Windows
   :synopsis: Module to abstract the data feature responses for wunderground

.. moduleauthor:: Tyler Santos <1tsantos at gmail.com>

"""
__docformat__ = 'reStructuredText'

# builtin P&Ms
from functools import reduce
from operator import getitem

# 3rd party P&Ms
from easydict import EasyDict


class WeatherBase():
    """Wrapper for one all data feature responses.

      Attributes:
          data (`EasyDict`): dictionary that allows for 'dotted' key references
          __data (dict): python dict, JSON representation of portion of response of interest
          NAN (list): list of values that are considered to be no data N/A type values
    """
    NAN = [-999, -9999]

    def __init__(self, data):
        """constructor to set up extract defaults for wunderground connection

        Args:
            data (dict): Interesting portion of wunderground's response

        Attributes:
            data (`EasyDict`): dictionary that allows for 'dotted' key references
            __data (dict): python dict, JSON representation of portion of response of interest

        """
        self.data = EasyDict(data)
        self.__data = data

    def extract_value(self, keys):
        """constructor to interface with feature response

        Args:
            keys (list): list of keys to drill down into nested dictionaries

        Returns:
            value of interest or None

        """

        if self.__data:
            try:
                return reduce(getitem, keys, self.__data)
            except KeyError:
                pass
            else:
                return None



"""
.. module:: weather
   :platform: Unix, Windows
   :synopsis: The classes and modules provided have several functions to aid in abstracting the nuances with the
                wunderground responses. The purpose is to expose the results in a uniform format while providing
                shortcuts, normalizing access, and making more pythonic references to wunderground's data.
.. moduleauthor:: Tyler Santos <1tsantos at gmail.com>

"""
__docformat__ = 'reStructuredText'

# 3rd party P&Ms
import requests

# local Ms
import WunderWeather
from WunderWeather import today, date, forecast, weather_base


class Extract():
    """Encapsulate logic for extracting weather data

    This is the main point of entry to extract data from the weather underground service
    utilizing their public API.
    
    Notes:
        `Wunderground Doc <https://www.wunderground.com/weather/api/d/docs>`_
        
        URL Request Format:
            `http://api.wunderground.com/api/<API_KEY>/features/settings/q/query.format`
    
    Attributes:
        BASE_URL (str): Base string used for URL generation
        FEATURE_URL (str): string template to generate URL for a feature request
        HURRICANE_URL (str): string template to generate URL for a hurricane feature request
        FEATURE_URL_MAP (dict): Mapping of module's feature key to wunderground's key in the URL
        FEATURE_RESPONSE_MAP (dict): Mapping of module's feature key to wunderground's key in the response
        FEATURE_CLASS_MAP (dict): Mapping of module's feature key to the object definition to generate an instance
    """
    BASE_URL = 'http://api.wunderground.com/api'
    FEATURE_URL = BASE_URL + '/{key}/{features}/{settings}/q/{query}.{format}'
    HURRICANE_URL = BASE_URL + '/{key}/{feature}/{settings}/view.{format}'

    FEATURE_URL_MAP = {
                        'alerts': 'alerts',
                        'astronomy': 'astronomy',
                        'cams': 'webcams',
                        'date': 'history_',
                        'date_range': 'planner_',
                        'daycast': 'forecast',
                        'geolookup': 'geolookup',
                        'hourly_daycast': 'hourly',
                        'hourly_weekcast': 'hourly10day',
                        'hurricane': 'currenthurricane',
                        'now': 'conditions',
                        'rawtide': 'rawtide',
                        'satelite': 'satelite',
                        'tide': 'tide',
                        'today_historical': 'almanac',
                        'weekcast': 'forecast10day',
                        'yesterday': 'yesterday',
                        }

    FEATURE_RESPONSE_MAP = {
                            'alerts': 'alerts',
                            'astronomy': 'moon_phase',
                            'cams': 'webcams',
                            'date': 'history',
                            'date_range': 'trip',
                            'daycast': 'forecast',
                            'geolookup': 'location',
                            'hourly_daycast': 'hourly_forecast',
                            'hourly_weekcast': 'hourly_forecast',
                            'hurricane': 'currenthurricane',
                            'now': 'current_observation',
                            'rawtide': 'rawtide',
                            'satelite': 'satelite',
                            'tide': 'tide',
                            'today_historical': 'almanac',
                            'weekcast': 'forecast',
                            'yesterday': 'history',
                            }

    FEATURE_CLASS_MAP = {
                            'alerts': 'weather_base.WeatherBase',
                            'astronomy': 'weather_base.WeatherBase',
                            'cams': 'weather_base.WeatherBase',
                            'date': 'date.Date',
                            'date_range': 'date.Range',
                            'daycast': 'forecast.Forecast',
                            'geolookup': 'weather_base.WeatherBase',
                            'hourly_daycast': 'weather_base.WeatherBase',
                            'hourly_weekcast': 'weather_base.WeatherBase',
                            'hurricane': 'weather_base.WeatherBase',
                            'now': 'today.Now',
                            'rawtide': 'weather_base.WeatherBase',
                            'satelite': 'weather_base.WeatherBase',
                            'tide': 'weather_base.WeatherBase',
                            'today_historical': 'today.Historical',
                            'weekcast': 'forecast.Forecast',
                            'yesterday': 'date.Date',
                            }

    @classmethod
    def get_feature_class(cls, feature_key):
        """Get the class definition for a particular feature

        Args:
            feature_key (str): internal key for feature

        Returns:
            Class definition for feature

        """
        pkg = 'WunderWeather'
        klass = cls.FEATURE_CLASS_MAP[feature_key]
        [mod,kls] = klass.split('.')
        mod = __import__(pkg+'.'+mod, fromlist=[kls])
        return getattr(mod, kls)

    def __init__(self, api_key, settings=None):
        """constructor to set up extract defaults for wunderground connection

        Args:
            api_key (str): Wunderground supplied API Key.
            settings (dict, optional): Settings for URL as defined.
                https://www.wunderground.com/weather/api/d/docs?d=data/index
                Defaults to lang:EN

        Attributes:
            api_key (str): Wunderground supplied API Key.
            format (str): Response format. currently only JSON is supported
            settings (dict): Settings for URL as defined.
                https://www.wunderground.com/weather/api/d/docs?d=data/index
                Defaults to {lang:EN}

        """
        self.api_key = api_key
        self.settings = settings if settings else {'lang': 'EN'}

        self.format = 'json'

    def hurricane(self):
        """Interface with the current hurricane data feature

        Attributes:
            feature_key (str): Module feature key
            context (dict): Used to populate feature URL template
            response (dict): JSON representation of the response
            response_feature_key (str): Module feature key's key for response parsing
            ctor (`weather.WeatherBase`): Reference to class to potentially generate an instance

        Returns:
            weather.weather_base instance or None

        """
        feature_key = 'hurricane'
        context = {
                    'key': self.api_key,
                    'feature': type(self).FEATURE_URL_MAP[feature_key],
                    'format': self.format,
                    'settings': '/'.join(['{0}:{1}'.format(skey, svalue) for skey, svalue in self.settings.items()]),
                    }
        response = self._make_feature_request(type(self).HURRICANE_URL, context)

        response_feature_key = type(self).FEATURE_RESPONSE_MAP[feature_key]

        if response_feature_key in response:
            ctor = type(self).get_feature_class(feature_key)
            return ctor({response_feature_key: response[response_feature_key]})

    def today_now(self, query):
        """Shorthand to interface with the conditions data feature

        Args:
            query (str or list): string or list of strings for query portion of URL generation

        Attributes:
            feature_context (tuple): tuple of tuples to give feature of interest
                and necessary data for that feature

        Returns:
            today.Now instance or None

        """
        feature_context = (('now', ''),)
        return self.features(query, feature_context)[0]

    def today_historical(self, query):
        """Shorthand to interface with the almanac data feature

        Args:
            query (str or list): string or list of strings for query portion of URL generation

        Attributes:
            feature_context (tuple): tuple of tuples to give feature of interest
                and necessary data for that feature

        Returns:
            today.Now instance or None

        """
        feature_context = (('today_historical', ''),)
        return self.features(query, feature_context)[0]

    def daycast(self, query):
        """Shorthand to interface with the forcast data feature

        Args:
            query (str or list): string or list of strings for query portion of URL generation

        Attributes:
            feature_context (tuple): tuple of tuples to give feature of interest
                and necessary data for that feature

        Returns:
            today.Now instance or None

        """
        feature_context = (('daycast', ''),)
        return self.features(query, feature_context)[0]

    def weekcast(self, query):
        """Shorthand to interface with the forcast10day data feature

        Args:
            query (str or list): string or list of strings for query portion of URL generation

        Attributes:
            feature_context (tuple): tuple of tuples to give feature of interest
                and necessary data for that feature

        Returns:
            today.Now instance or None

        """
        feature_context = (('weekcast', ''),)
        return self.features(query, feature_context)[0]

    def date(self, query, date):
        """Shorthand to interface with the history data feature

        Args:
            query (str or list): string or list of strings for query portion of URL generation
            date (str): Date in the form YYYYMMDD.

        Attributes:
            feature_context (tuple): tuple of tuples to give feature of interest
                and necessary data for that feature

        Returns:
            today.Now instance or None

        """
        feature_context = (('date', date),)
        return self.for_features(query, feature_context)[0]

    def date_range(self, query, date_range):
        """Shorthand to interface with the planner data feature

        Args:
            query (str or list): string or list of strings for query portion of URL generation
            date_range (str): Date range (30 day max) in the form MMDDMMDD.

        Attributes:
            feature_context (tuple): tuple of tuples to give feature of interest
            and necessary data for that feature

        Returns:
            today.Now instance or None

        """
        feature_context = (('date_range', date_range),)
        return self.features(query, feature_context)[0]

    def features(self, query, feature_context):
        """Central logic for making data feature requests

        Args:
            query (str or list): string or list of strings for query portion of URL generation
            feature_context (tuple): tuple of tuples to supply feature of interest
                and necessary data for that feature

        Attributes:
            query (list): Strings for URL generation
            feature_codes (str):  Final format for URL feature keys (with data appended)
            context (dict): final formatted data for URL template
            response (dict): JSON response
            ctor (obj): Object Class Reference to generate an instance.
                Could be one of

                * :mod:`weather_base.WeatherBase`
                * :mod:`today.Now`
                * :mod:`today.Historical`
                * :mod:`forcast.Daycast`
                * :mod:`forcast.Weekcast`
                * :mod:`date.Date`
                * :mod:`date.Range`

            weather_features (list): Instances to be retuned. Offset could be None
                if there was no response for the supplied data feature.
            feature_key (str): Module's feature key
            response_feature_key (str): Module's feature key for URL response

        Returns:
            List of weather_features. Offsets returned in the order they are supplied
            in the feature_context arg. Offset could be none if no response

        """
        if not type(query) == list:
            query = [query]

        feature_codes = [type(self).FEATURE_URL_MAP[f_key] + f_val for f_key,f_val in feature_context]

        context = {
                    'key': self.api_key,
                    'features': '/'.join(feature_codes),
                    'query': '/'.join(query),
                    'format': self.format,
                    'settings': '/'.join(['{0}:{1}'.format(s_key,s_val) for s_key,s_val in self.settings.items()]),
                    }
        response = self._make_feature_request(type(self).FEATURE_URL,context)

        weather_features = []
        for feature_key, _ in feature_context:
            
            ctor = type(self).get_feature_class(feature_key)

            response_feature_key = type(self).FEATURE_RESPONSE_MAP[feature_key]

            if response_feature_key in response:

                # if we get a list back as the high level item
                # keep the feature key in there as a single key dict
                if response[response_feature_key] is list:
                    weather_features.append(
                        ctor({response_feature_key: response[response_feature_key]}))
                else:
                    weather_features.append(
                        ctor(response[response_feature_key]))
            else:
                weather_features.append(None)

        return weather_features


    def _make_feature_request(self,url_template,data):
        """Private member to make a request 

        Args:
            url_template (str): Template used for URL generation
            data (dict): Data used to populate URL template

        Returns:
            JSONified response in dictionary format

        """
        return requests.get(url_template.format(**data)).json()
